# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/adv-ml',
                'doc_host': 'https://Irad-Zehavi.github.io',
                'git_url': 'https://github.com/Irad-Zehavi/adv-ml',
                'lib_path': 'adv_ml'},
  'syms': { 'adv_ml.all': {},
            'adv_ml.backdoor': { 'adv_ml.backdoor.BackdoorAttack': ('Core/backdoor.html#backdoorattack', 'adv_ml/backdoor.py'),
                                 'adv_ml.backdoor.BackdoorAttack._asr_dl': ( 'Core/backdoor.html#backdoorattack._asr_dl',
                                                                             'adv_ml/backdoor.py'),
                                 'adv_ml.backdoor.BackdoorAttack.after_create': ( 'Core/backdoor.html#backdoorattack.after_create',
                                                                                  'adv_ml/backdoor.py'),
                                 'adv_ml.backdoor.DataPoisoningAttack': ('Core/backdoor.html#datapoisoningattack', 'adv_ml/backdoor.py'),
                                 'adv_ml.backdoor.DataPoisoningAttack.__init__': ( 'Core/backdoor.html#datapoisoningattack.__init__',
                                                                                   'adv_ml/backdoor.py'),
                                 'adv_ml.backdoor.DataPoisoningAttack._poison': ( 'Core/backdoor.html#datapoisoningattack._poison',
                                                                                  'adv_ml/backdoor.py'),
                                 'adv_ml.backdoor.DataPoisoningAttack._poison_train_dl': ( 'Core/backdoor.html#datapoisoningattack._poison_train_dl',
                                                                                           'adv_ml/backdoor.py'),
                                 'adv_ml.backdoor.DataPoisoningAttack._subset_to_poison': ( 'Core/backdoor.html#datapoisoningattack._subset_to_poison',
                                                                                            'adv_ml/backdoor.py'),
                                 'adv_ml.backdoor.DataPoisoningAttack.after_create': ( 'Core/backdoor.html#datapoisoningattack.after_create',
                                                                                       'adv_ml/backdoor.py'),
                                 'adv_ml.backdoor._poisoned_dataset': ('Core/backdoor.html#_poisoned_dataset', 'adv_ml/backdoor.py')},
            'adv_ml.badnets': { 'adv_ml.badnets.BadNetsAttack': ('badnets.html#badnetsattack', 'adv_ml/badnets.py'),
                                'adv_ml.badnets.BadNetsAttack.__init__': ('badnets.html#badnetsattack.__init__', 'adv_ml/badnets.py'),
                                'adv_ml.badnets.BadNetsAttack._asr_dl': ('badnets.html#badnetsattack._asr_dl', 'adv_ml/badnets.py'),
                                'adv_ml.badnets.BadNetsAttack._insert_trigger': ( 'badnets.html#badnetsattack._insert_trigger',
                                                                                  'adv_ml/badnets.py'),
                                'adv_ml.badnets.BadNetsAttack._poison': ('badnets.html#badnetsattack._poison', 'adv_ml/badnets.py')},
            'adv_ml.evasion': { 'adv_ml.evasion.L2PGD': ('evasion.html#l2pgd', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.L2PGD.project_pert': ('evasion.html#l2pgd.project_pert', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.L2PGD.rand_init': ('evasion.html#l2pgd.rand_init', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.L2PGD.steepest_descent': ('evasion.html#l2pgd.steepest_descent', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.LinfPGD': ('evasion.html#linfpgd', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.LinfPGD.project_pert': ('evasion.html#linfpgd.project_pert', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.LinfPGD.rand_init': ('evasion.html#linfpgd.rand_init', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.LinfPGD.steepest_descent': ('evasion.html#linfpgd.steepest_descent', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.PGDCallback': ('evasion.html#pgdcallback', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.PGDCallback.__init__': ('evasion.html#pgdcallback.__init__', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.PGDCallback.after_batch': ('evasion.html#pgdcallback.after_batch', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.PGDCallback.before_step': ('evasion.html#pgdcallback.before_step', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.PGDCallback.clamp_pixel_values': ( 'evasion.html#pgdcallback.clamp_pixel_values',
                                                                                   'adv_ml/evasion.py'),
                                'adv_ml.evasion.PGDCallback.init_pert': ('evasion.html#pgdcallback.init_pert', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.PGDCallback.project_pert': ('evasion.html#pgdcallback.project_pert', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.PGDCallback.rand_init': ('evasion.html#pgdcallback.rand_init', 'adv_ml/evasion.py'),
                                'adv_ml.evasion.PGDCallback.steepest_descent': ( 'evasion.html#pgdcallback.steepest_descent',
                                                                                 'adv_ml/evasion.py'),
                                'adv_ml.evasion.PGDCallback.suggest_lr': ('evasion.html#pgdcallback.suggest_lr', 'adv_ml/evasion.py'),
                                'adv_ml.evasion._batch_norm': ('evasion.html#_batch_norm', 'adv_ml/evasion.py'),
                                'adv_ml.evasion._batch_normalize': ('evasion.html#_batch_normalize', 'adv_ml/evasion.py'),
                                'adv_ml.evasion._random_unit_vector': ('evasion.html#_random_unit_vector', 'adv_ml/evasion.py')},
            'adv_ml.input_optimization': { 'adv_ml.input_optimization.InputOptimizer': ( 'Core/input_optimization.html#inputoptimizer',
                                                                                         'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization.InputOptimizer.__init__': ( 'Core/input_optimization.html#inputoptimizer.__init__',
                                                                                                  'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization.InputOptimizer._perturb': ( 'Core/input_optimization.html#inputoptimizer._perturb',
                                                                                                  'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization.InputOptimizer.perturb': ( 'Core/input_optimization.html#inputoptimizer.perturb',
                                                                                                 'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization.PerturbationCallback': ( 'Core/input_optimization.html#perturbationcallback',
                                                                                               'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization.PerturbationCallback.init_pert': ( 'Core/input_optimization.html#perturbationcallback.init_pert',
                                                                                                         'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization.PerturbationCallback.suggest_lr': ( 'Core/input_optimization.html#perturbationcallback.suggest_lr',
                                                                                                          'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization._Perturbation': ( 'Core/input_optimization.html#_perturbation',
                                                                                        'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization._Perturbation.__init__': ( 'Core/input_optimization.html#_perturbation.__init__',
                                                                                                 'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization._Perturbation.forward': ( 'Core/input_optimization.html#_perturbation.forward',
                                                                                                'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization._TrainLoop': ( 'Core/input_optimization.html#_trainloop',
                                                                                     'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization._TrainLoop.before_train': ( 'Core/input_optimization.html#_trainloop.before_train',
                                                                                                  'adv_ml/input_optimization.py'),
                                           'adv_ml.input_optimization._TrainLoop.before_validate': ( 'Core/input_optimization.html#_trainloop.before_validate',
                                                                                                     'adv_ml/input_optimization.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/evasion.ipynb.

# %% auto 0
__all__ = ['PGDCallback', 'LinfPGD', 'L2PGD']

# %% ../nbs/evasion.ipynb 5
from abc import abstractmethod

import torch
from torch import nn
from fastai.vision.all import *

from .input_optimization import *


class PGDCallback(PerturbationCallback):
    "Implementes Projected Gradient Descent by bounding some $l_p$ norm of the perturbation"
    def __init__(self, epsilon=.3, rand_init=True):
        super().__init__()
        self.epsilon = epsilon
        self.p = nn.Parameter()

    def init_pert(self, x):
        self.p.data = self.rand_init(x.shape) if self.rand_init else torch.zeros(x.shape)
        self.clamp_pixel_values(x)
    
    def clamp_pixel_values(self, x):
        "Clamp peturbation such that perturbing `x` maintains valid pixel values"
        with torch.no_grad():
            x = x.to(self.p.device)
            self.p.data = (x + self.p).clamp(0., 1.) - x 

    def suggest_lr(self, input_opt: InputOptimizer):
        return self.epsilon / input_opt.epoch_size

    def before_step(self):
        with torch.no_grad():
            self.steepest_descent()

    def after_batch(self):
        with torch.no_grad():
            self.project_pert()
            self.clamp_pixel_values(self.x)

# %% ../nbs/evasion.ipynb 6
@patch()
@abstractmethod
def rand_init(self: PGDCallback, shape) -> Tensor:
    "Initialize a random perturbation in the $\epsilon$-ball"
    ...

# %% ../nbs/evasion.ipynb 7
@patch
@abstractmethod
def steepest_descent(self: PGDCallback) -> None:
    "Edit the perturbation's gradient to implement steepest descent"
    ...

# %% ../nbs/evasion.ipynb 8
@patch
@abstractmethod
def project_pert(self: PGDCallback) -> None:
    "Project the perturbation to the $\epsilon$-ball"
    ...

# %% ../nbs/evasion.ipynb 20
class LinfPGD(PGDCallback):
    "Implements PGD by bounding the $l_\infty$ norm"
    def rand_init(self, shape):
        return torch.rand(shape) * self.epsilon

    def steepest_descent(self):
        self.p.grad.sign_()

    def project_pert(self):
        self.p.clamp_(-self.epsilon, self.epsilon)

# %% ../nbs/evasion.ipynb 30
import torch.nn.functional as F
from torch.linalg import vector_norm


@delegates(F.normalize)
def _batch_normalize(t, **kwargs):
    return F.normalize(t.view(t.shape[0], -1), **kwargs).view(*t.shape)


@delegates(_batch_normalize)
def _random_unit_vector(shape, **kwargs):
    return _batch_normalize(torch.randn(shape), **kwargs)


def _batch_norm(t, **kwargs):
    return vector_norm(t.view(t.size(0), -1), **kwargs)


class L2PGD(PGDCallback):
    "Implements PGD by bounding the $l_2$ norm"
    def rand_init(self, shape):
        rand_dir = _random_unit_vector(shape)
        rand_len = torch.rand(1) * self.epsilon
        return rand_dir * rand_len

    def steepest_descent(self):
        self.p.grad.data = _batch_normalize(self.p.grad)

    def project_pert(self):
        norm = _batch_norm(self.p)
        self.p.mul_(torch.min(self.epsilon/norm, torch.ones_like(norm)))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/Core/backdoor.ipynb.

# %% auto 0
__all__ = ['BackdoorAttack', 'DataPoisoningAttack']

# %% ../nbs/Core/backdoor.ipynb 3
from abc import ABC, abstractmethod
from typing import Dict

from fastai.vision.all import *


class BackdoorAttack(ABC, Callback):
    "A `Callback` that affects the training process to install a backdoor. Also allows the measuring of the attack's success"
    def after_create(self):
         test_eq(len(self.dls), 2)
         self.dls.loaders.append(self._asr_dl())

    @abstractmethod
    def _asr_dl(self) -> DataLoader:
         "Returns a `DataLoader` used to measure the ASR (attack success rate)"
         ...

# %% ../nbs/Core/backdoor.ipynb 5
def _poisoned_dataset(clean, poison):
        all = clean + poison
        all.clean, all.poison = clean, poison
        if hasattr(clean, 'loss_func'):
            all.loss_func = clean.loss_func
        return all


class DataPoisoningAttack(BackdoorAttack):
    "A `BackdoorAttack` that installs the backdoor by altering a small portion of the training dataset"
    def __init__(self, test_only=False, poison_fraction=.1):
         super().__init__()
         store_attr('test_only, poison_fraction')

    def after_create(self):
         super().after_create()
         if not self.test_only:
            self._poison_train_dl()

    def _poison_train_dl(self):
        poison_size = int(self.poison_fraction * len(self.dls.train_ds))
        to_be_poisoned = self._subset_to_poison(self.dls.train_ds, poison_size)
        self.dls.train.dataset = _poisoned_dataset(clean=self.dls.train_ds - to_be_poisoned,
                                                   poison=self._poison(to_be_poisoned))

    def _subset_to_poison(self, clean_train_dataset: Datasets, size: int) -> Datasets:
        return clean_train_dataset.random_sub_dsets(size)

    @abstractmethod
    def _poison(self,
                data_to_poison: Datasets  # The prtion of the clean training data that will be replaced by poison. Could be used to construct the poison data
                ) -> Datasets:  # A dataset of poison data
        ...



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/badnets.ipynb.

# %% auto 0
__all__ = ['BadNetsAttack']

# %% ../nbs/badnets.ipynb 3
from fastai.vision.all import *

from .backdoor import *

@delegates()
class BadNetsAttack(DataPoisoningAttack):
    "https://arxiv.org/abs/1708.06733"
    def __init__(self,
                 trigger: np.array,  # Expected to be the same shape as the input
                 target_class: str,  # The target class for inputs with the trigger
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        store_attr('trigger, target_class')

    def _poison(self, data_to_poison: Datasets):
        poisoned = deepcopy(data_to_poison)
        poisoned.tls[0].tfms.add(mk_transform(self._insert_trigger))
        poisoned.tls[1].tfms = Pipeline([lambda _: self.target_class, data_to_poison.categorize])

        return poisoned

    def _insert_trigger(self, img):
        patched_np = (np.array(img)+self.trigger) % 256
        return type(img).create(patched_np)

    def _asr_dl(self):
        return self._poison(self.dls.valid_ds).dl()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/Core/input_optimization.ipynb.

# %% auto 0
__all__ = ['PerturbationCallback', 'InputOptimizer']

# %% ../nbs/Core/input_optimization.ipynb 4
from typing import Type
from abc import ABC, abstractmethod

from torch import nn
from fastai.vision.all import *


class PerturbationCallback(ABC, Callback):
    "Manages the input perturbation for an `InputOptimizer`"
    @abstractmethod
    def init_pert(self, x):
        ...
        
    @abstractmethod
    def suggest_lr(self, input_opt):
        ...
    

class _Perturbation(nn.Module):
    def __init__(self, p):
        super().__init__()
        self.p = p
        
    def forward(self, x):
        return x + self.p

                
class _TrainLoop(TrainEvalCallback):
    def before_train(self):
        super().before_train()
        self.model.eval()

    def before_validate(self):
        raise CancelValidException


class InputOptimizer(object):
    "Constructs adversarial examples: slightly perturbed inputs that fool classification models"
    def __init__(self,
                 model: Module,
                 pert_cb: Type[PerturbationCallback],
                 loss: Callable = CrossEntropyLossFlat(),
                 lr: float = None,  # pass `None` to try pick `lr` based on other parameters
                 targeted: bool = False,  # Whether the constructed inputs should be classified as the specified targets or not
                 min_delta: float = 1e-2,  # Minimum loss delta for `ReduceLROnPlateau` and `EarlyStoppingCallback`
                 min_lr: float = 1e-6,  # Minimum lr for `ReduceLROnPlateau`
                 # defaults taken from advertorch
                 epoch_size: int = 10,  # Affects how often epoch-callbacks are called (e.g. `Recorder`` and `EarlyStoppingCallback`)
                 n_epochs: int = 4):
        self.loss = loss if targeted else (lambda *args, **kwargs: -loss(*args, **kwargs))
        store_attr('model, pert_cb, lr, min_delta, min_lr, epoch_size, n_epochs')
        
        if self.lr is None:
            self.lr = pert_cb.suggest_lr(self)

        self.model.eval()
        self.model.requires_grad_(False)

    def perturb(self, dsets):
        x, y = dsets.load()
        x, y = x.detach().clone(), y.detach().clone()  # TODO: can I get rid of this?

        x_hat = self._perturb(x, y)

        return Datasets(tls=[TfmdLists(x_hat, ToTensor()),  # ToTensor for decoding
                             dsets.tls[1]])

    def _perturb(self, x, y):
        self.pert_cb.init_pert(x)

        learner = Learner(DataLoaders([(x, y) for _ in range(self.epoch_size)], []),
                          nn.Sequential(_Perturbation(self.pert_cb.p), self.model),
                          self.loss,
                          SGD,
                          self.lr,
                          train_bn=False,
                          default_cbs=False,
                          cbs=[_TrainLoop, Recorder(valid_metrics=False), ProgressCallback, BnFreeze,
                               self.pert_cb,
                               ReduceLROnPlateau('train_loss', min_delta=self.min_delta, min_lr=self.min_lr),
                               EarlyStoppingCallback('train_loss', min_delta=self.min_delta / 10)
                               ])
        learner.fit(self.n_epochs)
        return x.cpu() + self.pert_cb.p.data.detach().cpu()
